/*
 * codegen.c
 *
 * Ullrich von Bassewitz, 08.06.1998
 */



#include <stdio.h>
#include <string.h>

#include "global.h"
#include "version.h"
#include "error.h"
#include "check.h"
#include "io.h"
#include "mem.h"
#include "util.h"
#include "litpool.h"
#include "optimize.h"
#include "codegen.h"



/*****************************************************************************/
/*				     Data				     */
/*****************************************************************************/



/* Compiler relative stk ptr */
int oursp	= 0;

/* Current segment */
static enum { SEG_CODE, SEG_RODATA, SEG_DATA, SEG_BSS } CurSeg = SEG_CODE;

/* Segment names */
static char* SegmentNames [4];
static char* SegmentHints [4] = {
    "seg:code", "seg:rodata", "seg:data", "seg:bss"
};


/*****************************************************************************/
/*				    Helpers				     */
/*****************************************************************************/



static void typeerror (unsigned type)
/* Print an error message about an invalid operand type */
{
    Internal ("Invalid type in CF flags: %04X, type = %u", type, type & CF_TYPE);
}



static void CheckLocalOffs (unsigned Offs)
/* Check the offset into the stack for 8bit range */
{
    if (Offs >= 256) {
	/* Too many local vars */
	outline (";*** Too many locals");
	Error (ERR_TOO_MANY_LOCALS);
    }
}



static char* GetLabelName (unsigned flags, unsigned label, unsigned offs)
{
    static char lbuf [128];		/* Label name */

    /* Create the correct label name */
    switch (flags & CF_ADDRMASK) {

	case CF_STATIC:
	    /* Static memory cell */
	    sprintf (lbuf, "L%04X+%u", (unsigned)(label & 0xFFFF), offs);
	    break;

	case CF_EXTERNAL:
	    /* External label */
	    sprintf (lbuf, "_%s+%u", (char*) label, offs);
	    break;

	case CF_ABSOLUTE:
	    /* Absolute address */
	    sprintf (lbuf, "$%04X", (unsigned)((label+offs) & 0xFFFF));
	    break;

	case CF_REGVAR:
	    /* Variable in register bank */
	    sprintf (lbuf, "regbank+%u", (unsigned)((label+offs) & 0xFFFF));
	    break;

	default:
	    Internal ("Invalid address flags");
    }

    /* Return a pointer to the static buffer */
    return lbuf;
}



/*****************************************************************************/
/*			      Pre- and postamble			     */
/*****************************************************************************/



void g_preamble (void)
/* Generate the assembler code preamble */
{
    out ("; File generated by cc65 v %u.%u.%u\n\n", VER_MAJOR, VER_MINOR, VER_PATCH);

    /* Insert some object file options */
    out	(".fopt\t\tcompiler,\"cc65 v %u.%u.%u\"\n\n", VER_MAJOR, VER_MINOR, VER_PATCH);

    /* Allow auto import for runtime library routines */
    outstr (".autoimport\t+\n");

    /* Switch the assembler into case sensible mode */
    outstr (".case\t\t+\n");

    /* Tell the assembler if we want to generate debug info */
    out (".debuginfo\t%c\n", DebugInfo? '+' : '-');

    /* Import the stack pointer for direct auto variable access */
    outstr (".importzp\tsp, sreg, regsave, regbank, tmp1, ptr1\n");

    /* Define long branch macros */
    outstr (".macpack\tlongbranch\n\n");

    /* Define the ldax macro */
    out (".macro  ldax    Value\n"
    	 "        lda     #<(Value)\n"
         "        ldx     #>(Value)\n"
    	 ".endmacro\n\n");

    /* Flush the output queue */
    flushout ();

    /* Define the default names for the segments */
    SegmentNames [SEG_CODE] 	= xstrdup ("CODE");
    SegmentNames [SEG_RODATA]	= xstrdup ("RODATA");
    SegmentNames [SEG_DATA]	= xstrdup ("DATA");
    SegmentNames [SEG_BSS]	= xstrdup ("BSS");

    /* Tell the optimizer that this is the end of the preamble */
    OptAddHint ("end_of_preamble");
}



void g_postamble (void)
/* Generate assembler code postamble */
{
    /* Tell the optimizer that this is the start of the postamble */
    OptAddHint ("start_of_postamble");
}



/*****************************************************************************/
/*				Segment support				     */
/*****************************************************************************/



static void UseSeg (unsigned NewSeg)
/* Switch to a specific segment */
{
    if (CurSeg != NewSeg) {
  	CurSeg = NewSeg;
  	out (".segment\t\"%s\"\n", SegmentNames [CurSeg]);
	OptAddHint (SegmentHints [CurSeg]);
    }
}



void g_usecode (void)
/* Switch to the code segment */
{
    UseSeg (SEG_CODE);
}



void g_userodata (void)
/* Switch to the read only data segment */
{
    UseSeg (SEG_RODATA);
}



void g_usedata (void)
/* Switch to the data segment */
{
    UseSeg (SEG_DATA);
}



void g_usebss (void)
/* Switch to the bss segment */
{
    UseSeg (SEG_BSS);
}



void SegName (unsigned Seg, const char* Name)
/* Set the name of a segment */
{
    /* Free the old name and set a new one */
    xfree (SegmentNames [Seg]);
    SegmentNames [Seg] = xstrdup (Name);

    /* If the new segment is the current segment, emit a segment directive
     * with the new name.
     */
    if (Seg == CurSeg) {
    	CurSeg = -1;	/* Invalidate */
	UseSeg (Seg);
    }
}



void g_codename (const char* Name)
/* Set the name of the CODE segment */
{
    SegName (SEG_CODE, Name);
}



void g_rodataname (const char* Name)
/* Set the name of the RODATA segment */
{
    SegName (SEG_RODATA, Name);
}



void g_dataname (const char* Name)
/* Set the name of the DATA segment */
{
    SegName (SEG_DATA, Name);
}



void g_bssname (const char* Name)
/* Set the name of the BSS segment */
{
    SegName (SEG_BSS, Name);
}



/*****************************************************************************/
/*  			 	     Code				     */
/*****************************************************************************/



unsigned sizeofarg (unsigned flags)
/* Return the size of a function argument type that is encoded in flags */
{
    switch (flags & CF_TYPE) {

	case CF_CHAR:
	    return (flags & CF_FORCECHAR)? 1 : 2;

	case CF_INT:
	    return 2;

	case CF_LONG:
	    return 4;

	default:
	    typeerror (flags);
	    /* NOTREACHED */
	    return 2;
    }
}



int pop (unsigned flags)
/* Pop an argument of the given size */
{
    return oursp += sizeofarg (flags);
}



int push (unsigned flags)
/* Push an argument of the given size */
{
    return oursp -= sizeofarg (flags);
}



static void oljsr (char *str)
{
    outstr ("\tjsr\t");
    outline (str);
}



static unsigned MakeByteOffs (unsigned Flags, unsigned Offs)
/* The value in Offs is an offset to an address in a/x. Make sure, an object
 * of the type given in Flags can be loaded or stored into this address by
 * adding part of the offset to the address in ax, so that the remaining
 * offset fits into an index register. Return the remaining offset.
 */
{
    /* If the offset is too large for a byte register, add the high byte
     * of the offset to the primary. Beware: We need a special correction
     * if the offset in the low byte will overflow in the operation.
     */
    unsigned O = Offs & ~0xFFU;
    if ((Offs & 0xFF) > 256 - sizeofarg (Flags)) {
	/* We need to add the low byte also */
	O += Offs & 0xFF;
    }

    /* Do the correction if we need one */
    if (O) {
     	g_inc (CF_INT | CF_CONST, O);
     	Offs -= O;
    }

    /* Return the new offset */
    return Offs;
}



/*****************************************************************************/
/*		  	Functions handling local labels			     */
/*****************************************************************************/



void g_refloclabel (unsigned label, unsigned offset)
/* Generate a local label reference */
{
    outstr (GetLabelName (CF_STATIC, label, offset));
}



void g_defloclabel (unsigned label)
/* Define a local label */
{
    out ("L%04X:\n", label & 0xFFFF);
}



void g_refstrlabel (unsigned offset)
/* Reference a label that is somewhere in the string pool */
{
    out ("L%04X+%d", litlabel, offset);
}



/*****************************************************************************/
/*	     	       Functions handling global labels			     */
/*****************************************************************************/



void g_refregvar (unsigned label, unsigned offset)
/* Reference a register variable */
{
    outstr (GetLabelName (CF_REGVAR, label, offset));
}



void g_refgloblabel (unsigned label, unsigned offset)
/* Reference a global label */
{
    outstr (GetLabelName (CF_EXTERNAL, label, offset));
}



void g_defgloblabel (char* name)
/* Define a global label with the given name */
{
    out ("_%s:\n", name);
}



void g_defexport (char *sname, int ZP)
/* Export the given label */
{
    if (ZP) {
     	out ("\t.exportzp\t_%s\n", sname);
    } else {
     	out ("\t.export\t\t_%s\n", sname);
    }
}



void g_defimport (char *sname, int ZP)
/* Import the given label */
{
    if (ZP) {
       	out ("\t.importzp\t_%s\n", sname);
    } else {
     	out ("\t.import\t\t_%s\n", sname);
    }
}



/*****************************************************************************/
/*     		     Load functions for various registers		     */
/*****************************************************************************/



static void ldaconst (unsigned val)
/* Load a with a constant */
{
    out ("\tlda\t#$%02X\n", val & 0xFF);
}



static void ldxconst (unsigned val)
/* Load x with a constant */
{
    out ("\tldx\t#$%02X\n", val & 0xFF);
}



static void ldyconst (unsigned val)
/* Load y with a constant */
{
    out ("\tldy\t#$%02X\n", val & 0xFF);
}



/*****************************************************************************/
/*     			    Function entry and exit			     */
/*****************************************************************************/



/* Remember the argument size of a function. The variable is set by g_enter
 * and used by g_leave. If the functions gets its argument size by the caller
 * (variable param list or function without prototype), g_enter will set the
 * value to -1.
 */
static int funcargs;


void g_enter (unsigned flags, char *name, unsigned argsize)
/* Function prologue */
{
    g_defgloblabel (name);    	/* Define function name as label */
    if (flags & CF_FIXARGC) {
	/* Just remember the argument size for the leave */
	funcargs = argsize;
    } else {
       	funcargs = -1;
       	out ("\tjsr\tenter\n");
    }
}



void g_leave (int flags, int val)
/* Function epilogue */
{
    int k;
    char buf [40];

    /* How many bytes of locals do we have to drop? */
    k = -oursp;

    /* If we didn't have a variable argument list, don't call leave */
    if (funcargs >= 0) {

	/* Load a function return code if needed */
	if (flags & CF_CONST) {
	    g_getimmed (flags, val);
	}

	/* Drop stackframe or leave with rts */
	k += funcargs;
	if (k == 0) {
	    out ("\trts\n");
	} else if (k <= 8) {
	    out ("\tjmp\tincsp%d\n", k);
	} else {
	    CheckLocalOffs (k);
	    ldyconst (k);
	    out ("\tjmp\taddysp\n");
	}

    } else {

     	strcpy (buf, "\tjmp\tleave");
     	if (k) {
     	    /* We've a stack frame to drop */
     	    ldyconst (k);
     	    strcat (buf, "y");
     	}
     	if (flags & CF_CONST) {
     	    if ((flags & CF_TYPE) != CF_LONG) {
     	   	/* Constant int sized value given for return code */
     	   	if (val == 0) {
     	   	    /* Special case: return 0 */
     	   	    strcat (buf, "00");
     	   	} else if (!((val >> 8) & 0xFF)) {
     	   	    /* Special case: constant with high byte zero */
     	   	    ldaconst (val);		/* Load low byte */
     		    strcat (buf, "0");
     		} else {
     		    /* Others: arbitrary constant value */
     		    g_getimmed (flags, val);	/* Load value */
     		}
     	    } else {
     		/* Constant long value: No shortcut possible */
     		g_getimmed (flags, val);
     	    }
     	}

     	/* Output the jump */
     	outline (buf);
    }
}



/*****************************************************************************/
/*		       	      Register variables			     */
/*****************************************************************************/



void g_save_regvars (int RegOffs, unsigned Bytes)
/* Save register variables */
{
    /* Don't loop for up to two bytes */
    if (Bytes == 1) {

     	out ("\tlda\tregbank%+d\n"
     	     "\tjsr\tpusha\n",
     	     RegOffs);

    } else if (Bytes == 2) {

     	out ("\tlda\tregbank%+d\n"
     	     "\tldx\tregbank%+d\n"
     	     "\tjsr\tpushax\n",
     	     RegOffs,
     	     RegOffs+1);

    } else {

     	/* More than two bytes - loop */
     	unsigned Label = getlabel ();
     	g_space (Bytes);
     	ldyconst (Bytes - 1);
     	ldxconst (Bytes);
     	g_defloclabel (Label);
     	out ("\tlda\tregbank%+d,x\n"
     	     "\tsta\t(sp),y\n"
     	     "\tdey\n"
     	     "\tdex\n"
     	     "\tbne\tL%04X\n",
     	     RegOffs-1,
     	     Label);

    }

    /* We pushed stuff, correct the stack pointer */
    oursp -= Bytes;
}



void g_restore_regvars (int StackOffs, int RegOffs, unsigned Bytes)
/* Restore register variables */
{
    /* Calculate the actual stack offset and check it */
    StackOffs -= oursp;
    CheckLocalOffs (StackOffs);

    /* Don't loop for up to two bytes */
    if (Bytes == 1) {

     	ldyconst (StackOffs);
     	out ("\tlda\t(sp),y\n"
     	     "\tsta\tregbank%+d\n",
     	     RegOffs);

    } else if (Bytes == 2) {

     	ldyconst (StackOffs);
     	out ("\tlda\t(sp),y\n"
     	     "\tsta\tregbank%+d\n"
       	     "\tiny\n"
	     "\tlda\t(sp),y\n"
	     "\tsta\tregbank%+d\n",
	     RegOffs,
	     RegOffs+1);

    } else {

     	/* More than two bytes - loop */
     	unsigned Label = getlabel ();
     	ldyconst (StackOffs+Bytes-1);
     	ldxconst (Bytes);
     	g_defloclabel (Label);
     	out ("\tlda\t(sp),y\n"
     	     "\tsta\tregbank%+d,x\n"
     	     "\tdey\n"
     	     "\tdex\n"
     	     "\tbne\tL%04X\n",
	     RegOffs-1,
	     Label);

    }
}



/*****************************************************************************/
/*			     Fetching memory cells	   		     */
/*****************************************************************************/



void g_getimmed (unsigned flags, unsigned long val)
/* Load a constant into the primary register, CF_CONST implied */
{
    switch (flags & CF_TYPE) {

   	case CF_CHAR:
   	    if (flags & CF_FORCECHAR) {
   	    	ldaconst (val);
   	    	break;
   	    }
   	    /* FALL THROUGH */
   	case CF_INT:
       	    ldxconst ((val >> 8) & 0xFF);
   	    ldaconst (val & 0xFF);
   	    break;

   	case CF_LONG:
   	    if (val < 0x100) {
   	  	out ("\tldx\t#$00\n"
   	  	     "\tstx\tsreg+1\n"
   	  	     "\tstx\tsreg\n"
   	  	     "\tlda\t#$%02X\n",
   	  	     (unsigned char) val);
   	    } else if ((val & 0xFFFF00FF) == 0) {
   	  	out ("\tlda\t#$00\n"
   	  	     "\tsta\tsreg+1\n"
   	  	     "\tsta\tsreg\n"
   	  	     "\tldx\t#$%02X\n",
   	  	     (unsigned char) (val >> 8));
   	    } else if ((val & 0xFFFF0000) == 0 && !FavourSize) {
   	  	out ("\tlda\t#$00\n"
   		     "\tsta\tsreg+1\n"
		     "\tsta\tsreg\n"
		     "\tlda\t#$%02X\n"
		     "\tldx\t#$%02X\n",
		     (unsigned char) val,
		     (unsigned char) (val >> 8));
	    } else if ((val & 0xFFFFFF00) == 0xFFFFFF00) {
		out ("\tldx\t#$FF\n"
   		     "\tstx\tsreg+1\n"
   		     "\tstx\tsreg\n");
		if ((val & 0xFF) == 0xFF) {
		    out ("\ttxa\n");
		} else {
		    out ("\tlda\t#$%02X\n",
			 (unsigned char) val);
		}
	    } else if ((val & 0xFFFF00FF) == 0xFFFF00FF) {
		out ("\tlda\t#$FF\n"
   		     "\tsta\tsreg+1\n"
   		     "\tsta\tsreg\n"
		     "\tldx\t#$%02X\n",
		     (unsigned char) (val >> 8));
	    } else {
	        oljsr ("ldeax"); 	/* Call subroutine... */
	        g_deflong (val); 	/* ...that will load following value */
	    }
	    break;

	default:
	    typeerror (flags);
	    break;

    }
}



void g_getstatic (unsigned flags, unsigned label, unsigned offs)
/* Fetch an static memory cell into the primary register */
{
    /* Create the correct label name */
    char* lbuf = GetLabelName (flags, label, offs);

    /* Check the size and generate the correct load operation */
    switch (flags & CF_TYPE) {

     	case CF_CHAR:
     	    if ((flags & CF_FORCECHAR) || (flags & CF_TEST)) {
     	        out ("\tlda\t%s\n", lbuf);	/* load A from the label */
       	    } else {
     	     	ldxconst (0);
     	     	out ("\tlda\t%s\n", lbuf);	/* load A from the label */
     	     	if (!(flags & CF_UNSIGNED)) {
     	     	    /* Must sign extend */
     	     	    out ("\tbpl\t*+3\n");
     	     	    out ("\tdex\n");
     		    OptAddHint ("x:!");		/* X is invalid now */
     	     	}
     	    }
     	    break;

     	case CF_INT:
     	    out ("\tlda\t%s\n", lbuf);
     	    if (flags & CF_TEST) {
     		out ("\tora\t%s+1\n", lbuf);
     	    } else {
     		out ("\tldx\t%s+1\n", lbuf);
     	    }
     	    break;

   	case CF_LONG:
	    if (flags & CF_TEST) {
	     	out ("\tlda\t%s+3\n"
		     "\tora\t%s+2\n"
		     "\tora\t%s+1\n"
		     "\tora\t%s\n",
		     lbuf, lbuf, lbuf, lbuf);
	    } else {
	     	out ("\tlda\t%s+3\n"
	     	     "\tsta\tsreg+1\n"
	     	     "\tlda\t%s+2\n"
	     	     "\tsta\tsreg\n"
       	     	     "\tldx\t%s+1\n"
	     	     "\tlda\t%s\n",
	     	     lbuf, lbuf, lbuf, lbuf);
	    }
	    break;

       	default:
       	    typeerror (flags);

    }
}



void g_getlocal (unsigned flags, int offs)
/* Fetch specified local object (local var). */
{
    offs -= oursp;
    CheckLocalOffs (offs);
    switch (flags & CF_TYPE) {

	case CF_CHAR:
	    if ((flags & CF_FORCECHAR) || (flags & CF_TEST)) {
		ldyconst (offs);
	    	outline ("\tlda\t(sp),y");
	    } else {
		if (offs == 0) {
		    outline ("\tldx\t#$00");
		    outline ("\tlda\t(sp,x)");
		} else {
		    ldyconst (offs);
		    outline ("\tldx\t#$00");
		    outline ("\tlda\t(sp),y");
		}
     	    	if ((flags & CF_UNSIGNED) == 0) {
     	    	    outline ("\tbpl\t*+3");
     	 	    outline ("\tdex");
     		    OptAddHint ("x:!");		/* X is invalid now */
	 	}
	    }
	    break;

	case CF_INT:
	    CheckLocalOffs (offs + 1);
       	    if (flags & CF_TEST) {
	    	ldyconst (offs + 1);
	    	out ("\tlda\t(sp),y\n"
	    	     "\tdey\n"
	    	     "\tora\t(sp),y\n");
	    } else {
		if (FavourSize) {
		    if (offs) {
			ldyconst (offs+1);
       			outline ("\tjsr\tldaxysp");
		    } else {
			outline ("\tjsr\tldax0sp");
		    }
		} else {
	    	    ldyconst (offs + 1);
		    out ("\tlda\t(sp),y\n"
		  	 "\ttax\n"
			 "\tdey\n"
			 "\tlda\t(sp),y\n");
		}
	    }
	    break;

	case CF_LONG:
    	    if (offs) {
    	 	ldyconst (offs+3);
    	 	outline ("\tjsr\tldeaxysp");
    	    } else {
    	 	outline ("\tjsr\tldeax0sp");
    	    }
	    break;

    	default:
    	    typeerror (flags);
    }
}



void g_getind (unsigned flags, unsigned offs)
/* Fetch the specified object type indirect through the primary register
 * into the primary register
 */
{
    /* If the offset is greater than 255, add the part that is > 255 to
     * the primary. This way we get an easy addition and use the low byte
     * as the offset
     */
    offs = MakeByteOffs (flags, offs);

    /* Handle the indirect fetch */
    switch (flags & CF_TYPE) {

     	case CF_CHAR:
       	    /* Character sized */
     	    if (offs) {
     		ldyconst (offs);
     	        if (flags & CF_UNSIGNED) {
     	     	    outline ("\tjsr\tldauidx");
       	     	} else {
     	     	    outline ("\tjsr\tldaidx");
     	     	}
     	    } else {
     	        if (flags & CF_UNSIGNED) {
     		    if (FavourSize) {
     	     	        outline ("\tjsr\tldaui");
     		    } else {
     			outline ("\tsta\tptr1");
     			outline ("\tstx\tptr1+1");
     		     	outline ("\tldx\t#$00");
     			outline ("\tlda\t(ptr1,x)");
     		    }
     	     	} else {
     	     	    outline ("\tjsr\tldai");
     	     	}
     	    }
     	    break;

     	case CF_INT:
     	    if (flags & CF_TEST) {
     		ldyconst (offs);
     		outline ("\tsta\tptr1");
     		outline ("\tstx\tptr1+1");
     		outline ("\tlda\t(ptr1),y");
     		outline ("\tiny");
     		outline ("\tora\t(ptr1),y");
     	    } else {
     		if (offs == 0) {
     		    outline ("\tjsr\tldaxi");
     		} else {
     		    ldyconst (offs+1);
     		    outline ("\tjsr\tldaxidx");
     		}
     	    }
     	    break;

       	case CF_LONG:
     	    if (offs == 0) {
     		outline ("\tjsr\tldeaxi");
     	    } else {
     		ldyconst (offs+3);
     		outline ("\tjsr\tldeaxidx");
     	    }
     	    if (flags & CF_TEST) {
       		outline ("\tjsr\ttsteax");
     	    }
     	    break;

     	default:
     	    typeerror (flags);

    }
}



void g_leasp (int offs)
/* Fetch the address of the specified symbol into the primary register */
{
    /* Calculate the offset relative to sp */
    offs -= oursp;

    /* For value 0 we do direct code */
    if (offs == 0) {
	outline ("\tlda\tsp");
	outline ("\tldx\tsp+1");
    } else {
	if (FavourSize) {
	    ldyconst (offs);       	/* Load Y with offset value */
	    oljsr ("leaysp");      	/* Load effective address */
	} else {
	    ldaconst (offs);
	    outline ("\tclc");
	    outline ("\tldx\tsp+1");
	    outline ("\tadc\tsp");
	    outline ("\tbcc\t*+3");
	    outline ("\tinx");
	    OptAddHint ("x:!");		/* Invalidate X */
	}
    }
}



/*****************************************************************************/
/*     	    		       Store into memory			     */
/*****************************************************************************/



void g_putstatic (unsigned flags, unsigned long label, unsigned offs)
/* Store the primary register into the specified static memory cell */
{
    /* Create the correct label name */
    char* lbuf = GetLabelName (flags, label, offs);

    /* Check the size and generate the correct store operation */
    switch (flags & CF_TYPE) {

     	case CF_CHAR:
    	    out ("\tsta\t%s\n", lbuf);
     	    break;

     	case CF_INT:
    	    out ("\tsta\t%s\n", lbuf);
	    out ("\tstx\t%s+1\n", lbuf);
     	    break;

     	case CF_LONG:
    	    out ("\tsta\t%s\n", lbuf);
	    out ("\tstx\t%s+1\n", lbuf);
	    out ("\tldy\tsreg\n");
	    out ("\tsty\t%s+2\n", lbuf);
	    out ("\tldy\tsreg+1\n");
	    out ("\tsty\t%s+3\n", lbuf);
     	    break;

       	default:
       	    typeerror (flags);

    }
}



void g_putlocal (unsigned flags, int offs)
/* Put data into local object. */
{
    offs -= oursp;
    CheckLocalOffs (offs);
    switch (flags & CF_TYPE) {

     	case CF_CHAR:
       	    ldyconst (offs);
     	    outline ("\tsta\t(sp),y");
     	    break;

     	case CF_INT:
     	    if (offs) {
     		ldyconst (offs);
     		oljsr ("staxysp");
     	    } else {
     		oljsr ("stax0sp");
     	    }
     	    break;

     	case CF_LONG:
     	    if (offs) {
     	     	ldyconst (offs);
     	     	oljsr ("steaxysp");
     	    } else {
     	     	oljsr ("steax0sp");
     	    }
     	    break;

       	default:
     	    typeerror (flags);

    }
}



void g_putind (unsigned flags, unsigned offs)
/* Store the specified object type in the primary register at the address
 * on the top of the stack
 */
{
    /* We cannot currently handle more than byte sized offsets */
    if (offs > 256 - sizeofarg (flags)) {
	Internal ("g_putind: Large offsets not implemented");
    }

    /* Check the size and determine operation */
    switch (flags & CF_TYPE) {

     	case CF_CHAR:
     	    if (offs) {
     	        ldyconst (offs);
     		outline ("\tjsr\tstaspidx");
     	    } else {
     		outline ("\tjsr\tstaspp");
     	    }
     	    break;

     	case CF_INT:
     	    if (offs) {
     	        ldyconst (offs);
     		outline ("\tjsr\tstaxspidx");
     	    } else {
     		outline ("\tjsr\tstaxspp");
     	    }
     	    break;

     	case CF_LONG:
     	    if (offs) {
     	        ldyconst (offs);
     		outline ("\tjsr\tsteaxspidx");
     	    } else {
     		outline ("\tjsr\tsteaxspp");
     	    }
     	    break;

     	default:
     	    typeerror (flags);

    }

    /* Pop the argument */
    pop (flags);
}



/*****************************************************************************/
/*		      type conversion and similiar stuff		     */
/*****************************************************************************/



void g_toslong (unsigned flags)
/* Make sure, the value on TOS is a long. Convert if necessary */
{
    switch (flags & CF_TYPE) {

	case CF_CHAR:
	case CF_INT:
	    if (flags & CF_UNSIGNED) {
		out ("\tjsr\ttosulong\n");
	    } else {
		out ("\tjsr\ttoslong\n");
	    }
	    push (CF_INT);
	    break;

	case CF_LONG:
	    break;

	default:
	    typeerror (flags);
    }
}



void g_tosint (unsigned flags)
/* Make sure, the value on TOS is an int. Convert if necessary */
{
    switch (flags & CF_TYPE) {

	case CF_CHAR:
	case CF_INT:
	    break;

	case CF_LONG:
	    out ("\tjsr\ttosint");
	    pop (CF_INT);
	    break;

	default:
	    typeerror (flags);
    }
}



void g_reglong (unsigned flags)
/* Make sure, the value in the primary register a long. Convert if necessary */
{
    switch (flags & CF_TYPE) {

	case CF_CHAR:
	case CF_INT:
	    if (flags & CF_UNSIGNED) {
		if (FavourSize) {
		    out ("\tjsr\taxulong\n");
		} else {
		    ldyconst (0);
		    out ("\tsty\tsreg\n\tsty\tsreg+1\n");
		}
	    } else {
		out ("\tjsr\taxlong\n");
	    }
	    break;

	case CF_LONG:
	    break;

	default:
	    typeerror (flags);
    }
}



unsigned g_typeadjust (unsigned lhs, unsigned rhs)
/* Adjust the integer operands before doing a binary operation. lhs is a flags
 * value, that corresponds to the value on TOS, rhs corresponds to the value
 * in (e)ax. The return value is the the flags value for the resulting type.
 */
{
    unsigned ltype, rtype;
    unsigned result;

    /* Get the type spec from the flags */
    ltype = lhs & CF_TYPE;
    rtype = rhs & CF_TYPE;

    /* Check if a conversion is needed */
    if (ltype == CF_LONG && rtype != CF_LONG && (rhs & CF_CONST) == 0) {
   	/* We must promote the primary register to long */
   	g_reglong (rhs);
   	/* Get the new rhs type */
   	rhs = (rhs & ~CF_TYPE) | CF_LONG;
   	rtype = CF_LONG;
    } else if (ltype != CF_LONG && (lhs & CF_CONST) == 0 && rtype == CF_LONG) {
   	/* We must promote the lhs to long */
	if (lhs & CF_REG) {
	    g_reglong (lhs);
	} else {
   	    g_toslong (lhs);
	}
   	/* Get the new rhs type */
   	lhs = (lhs & ~CF_TYPE) | CF_LONG;
   	ltype = CF_LONG;
    }

    /* Determine the result type for the operation:
     *	- The result is const if both operands are const.
     *	- The result is unsigned if one of the operands is unsigned.
     *	- The result is long if one of the operands is long.
     *	- Otherwise the result is int sized.
     */
    result = (lhs & CF_CONST) & (rhs & CF_CONST);
    result |= (lhs & CF_UNSIGNED) | (rhs & CF_UNSIGNED);
    if (rtype == CF_LONG || ltype == CF_LONG) {
	result |= CF_LONG;
    } else {
	result |= CF_INT;
    }
    return result;
}



unsigned g_typecast (unsigned lhs, unsigned rhs)
/* Cast the value in the primary register to the operand size that is flagged
 * by the lhs value. Return the result value.
 */
{
    unsigned ltype, rtype;

    /* Get the type spec from the flags */
    ltype = lhs & CF_TYPE;
    rtype = rhs & CF_TYPE;

    /* Check if a conversion is needed */
    if (ltype == CF_LONG && rtype != CF_LONG && (rhs & CF_CONST) == 0) {
	/* We must promote the primary register to long */
	g_reglong (rhs);
    }

    /* Do not need any other action. If the left type is int, and the primary
     * register is long, it will be automagically truncated. If the right hand
     * side is const, it is not located in the primary register and handled by
     * the expression parser code.
     */

    /* Result is const if the right hand side was const */
    lhs |= (rhs & CF_CONST);

    /* The resulting type is that of the left hand side (that's why you called
     * this function :-)
     */
    return lhs;
}



void g_scale (unsigned flags, long val)
/* Scale the value in the primary register by the given value. If val is positive,
 * scale up, is val is negative, scale down. This function is used to scale
 * the operands or results of pointer arithmetic by the size of the type, the
 * pointer points to.
 */
{
    int p2;

    /* Value may not be zero */
    if (val == 0) {
       	Internal ("Data type has no size");
    } else if (val > 0) {

     	/* Scale up */
     	if ((p2 = powerof2 (val)) > 0 && p2 <= 3) {

     	    /* Factor is 2, 4 or 8, use special function */
     	    switch (flags & CF_TYPE) {

     		case CF_CHAR:
     		    if (flags & CF_FORCECHAR) {
     		     	while (p2--) {
     		     	    outline ("\tasl\ta");
     		     	}
     		     	break;
     		    }
     		    /* FALLTHROUGH */

     		case CF_INT:
     		    if (FavourSize || p2 >= 3) {
     		       	if (flags & CF_UNSIGNED) {
     		     	    out ("\tjsr\tshlax%d\n", p2);
     		     	} else {
     		     	    out ("\tjsr\taslax%d\n", p2);
     		     	}
     		    } else {
     			outline ("\tstx\ttmp1");
     		  	while (p2--) {
     			    outline ("\tasl\ta\n"
     				     "\trol\ttmp1");
     			}
     			outline ("\tldx\ttmp1");
     		    }
     		    break;

     		case CF_LONG:
     		    if (flags & CF_UNSIGNED) {
     		     	out ("\tjsr\tshleax%d\n", p2);
     		    } else {
     			out ("\tjsr\tasleax%d\n", p2);
     		    }
     		    break;

     		default:
     		    typeerror (flags);

     	    }

     	} else if (val != 1) {

       	    /* Use a multiplication instead */
     	    g_mul (flags | CF_CONST, val);

     	}

    } else {

     	/* Scale down */
     	val = -val;
     	if ((p2 = powerof2 (val)) > 0 && p2 <= 3) {

     	    /* Factor is 2, 4 or 8, use special function */
     	    switch (flags & CF_TYPE) {

     		case CF_CHAR:
     		    if (flags & CF_FORCECHAR) {
     			if (flags & CF_UNSIGNED) {
     			    while (p2--) {
     			      	outline ("\tlsr\ta");
     			    }
     			    break;
     			} else if (p2 <= 2) {
     		  	    outline ("\tcmp\t#$80");
     			    outline ("\tror\ta");
     			    break;
     			}
     		    }
     		    /* FALLTHROUGH */

     		case CF_INT:
     		    if (flags & CF_UNSIGNED) {
			if (FavourSize || p2 >= 3) {
     			    out ("\tjsr\tlsrax%d\n", p2);
			} else {
			    outline ("\tstx\ttmp1");
			    while (p2--) {
				outline ("\tlsr\ttmp1\n"
					 "\tror\ta");
			    }
			    outline ("\tldx\ttmp1");
			}
     		    } else {
			if (FavourSize || p2 >= 3) {
     			    out ("\tjsr\tasrax%d\n", p2);
			} else {
			    outline ("\tstx\ttmp1");
			    while (p2--) {
				outline ("\tcpx\t#$80\n"
    					 "\tror\ttmp1\n"
					 "\tror\ta");
			    }
			    outline ("\tldx\ttmp1");
	    	     	}
     		    }
     		    break;

     		case CF_LONG:
     		    if (flags & CF_UNSIGNED) {
     		     	out ("\tjsr\tlsreax%d\n", p2);
     		    } else {
     		       	out ("\tjsr\tasreax%d\n, p2");
     		    }
     		    break;

     		default:
     		    typeerror (flags);

     	    }

     	} else if (val != 1) {

       	    /* Use a division instead */
     	    g_div (flags | CF_CONST, val);

     	}
    }
}



/*****************************************************************************/
/*		Adds and subs of variables fix a fixed address		     */
/*****************************************************************************/



void g_addlocal (unsigned flags, int offs)
/* Add a local variable to ax */
{
    /* Correct the offset and check it */
    offs -= oursp;
    CheckLocalOffs (offs);

    switch (flags & CF_TYPE) {

     	case CF_CHAR:
	    out ("\tldy\t#$%02X\n", offs & 0xFF);
	    out ("\tclc\n");
	    out ("\tadc\t(sp),y\n");
	    out ("\tbcc\t*+3\n");
	    out ("\tinx\n");
	    OptAddHint ("x:!");
	    break;

     	case CF_INT:
     	    out ("\tldy\t#$%02X\n", offs & 0xFF);
     	    out ("\tclc\n");
     	    out ("\tadc\t(sp),y\n");
     	    out ("\tpha\n");
     	    out ("\ttxa\n");
     	    out ("\tiny\n");
     	    out ("\tadc\t(sp),y\n");
     	    out ("\ttax\n");
     	    out ("\tpla\n");
     	    break;

     	case CF_LONG:
     	    /* Do it the old way */
       	    g_push (flags, 0);
     	    g_getlocal (flags, offs);
     	    g_add (flags, 0);
     	    break;

     	default:
     	    typeerror (flags);

    }
}



void g_addstatic (unsigned flags, unsigned label, unsigned offs)
/* Add a static variable to ax */
{
    /* Create the correct label name */
    char* lbuf = GetLabelName (flags, label, offs);

    switch (flags & CF_TYPE) {

	case CF_CHAR:
	    out ("\tclc\n");
	    out ("\tadc\t%s\n", lbuf);
	    out ("\tbcc\t*+3\n");
	    out ("\tinx\n");
	    OptAddHint ("x:!");
	    break;

	case CF_INT:
	    out ("\tclc\n");
	    out ("\tadc\t%s\n", lbuf);
	    out ("\ttay\n");
	    out ("\ttxa\n");
     	    out ("\tadc\t%s+1\n", lbuf);
	    out ("\ttax\n");
	    out ("\ttya\n");
	    break;

	case CF_LONG:
	    /* Do it the old way */
       	    g_push (flags, 0);
	    g_getstatic (flags, label, offs);
	    g_add (flags, 0);
	    break;

	default:
	    typeerror (flags);

    }
}



/*****************************************************************************/
/*	       Compares of ax with a variable with fixed address	     */
/*****************************************************************************/



void g_cmplocal (unsigned flags, int offs)
/* Compare a local variable to ax */
{
    Internal ("g_cmplocal not implemented");
}



void g_cmpstatic (unsigned flags, unsigned label, unsigned offs)
/* Compare a static variable to ax */
{
    Internal ("g_cmpstatic not implemented");
}



/*****************************************************************************/
/*   	  	       	     Special op= functions			     */
/*****************************************************************************/



void g_addeqstatic (unsigned flags, unsigned label, unsigned offs,
       		    unsigned long val)
/* Emit += for a static variable */
{
    /* Create the correct label name */
    char* lbuf = GetLabelName (flags, label, offs);

    /* Check the size and determine operation */
    switch (flags & CF_TYPE) {

       	case CF_CHAR:
       	    if (flags & CF_FORCECHAR) {
     		out ("\tldx\t#$00\n");
       		if (flags & CF_CONST) {
     		    if (val == 1) {
     		   	out ("\tinc\t%s\n", lbuf);
     		   	out ("\tlda\t%s\n", lbuf);
     		    } else {
       	       	       	out ("\tlda\t#$%02X\n", val & 0xFF);
     		   	out ("\tclc\n");
     		   	out ("\tadc\t%s\n", lbuf);
     			out ("\tsta\t%s\n", lbuf);
     		    }
       	       	} else {
     		    out ("\tclc\n");
       		    out ("\tadc\t%s\n", lbuf);
     		    out ("\tsta\t%s\n", lbuf);
       		}
     		if ((flags & CF_UNSIGNED) == 0) {
     		    out ("\tbpl\t*+3\n");
     		    out ("\tdex\n");
     		    OptAddHint ("x:!");	       	/* Invalidate X */
     		}
       		break;
       	    }
       	    /* FALLTHROUGH */

       	case CF_INT:
       	    if (flags & CF_CONST) {
     		if (val == 1) {
     		    label = getlabel ();
     		    out ("\tinc\t%s\n", lbuf);
     		    out ("\tbne\tL%04X\n", label);
     		    out ("\tinc\t%s+1\n", lbuf);
     		    g_defloclabel (label);
     		    out ("\tlda\t%s\n", lbuf);		/* Hmmm... */
     		    out ("\tldx\t%s+1\n", lbuf);
     		} else {
       	       	    out ("\tlda\t#$%02X\n", val & 0xFF);
     		    out ("\tclc\n");
     		    out ("\tadc\t%s\n", lbuf);
     		    out ("\tsta\t%s\n", lbuf);
     		    if (val < 0x100) {
     		       	label = getlabel ();
     		       	out ("\tbcc\tL%04X\n", label);
     		       	out ("\tinc\t%s+1\n", lbuf);
       		       	g_defloclabel (label);
     		       	out ("\tldx\t%s+1\n", lbuf);
     		    } else {
       	       	       	out ("\tlda\t#$%02X\n", (val >> 8) & 0xFF);
     		       	out ("\tadc\t%s+1\n", lbuf);
     		       	out ("\tsta\t%s+1\n", lbuf);
     		       	out ("\ttax\n");
     		       	out ("\tlda\t%s\n", lbuf);
     		    }
     		}
       	    } else {
     		out ("\tclc\n");
       		out ("\tadc\t%s\n", lbuf);
       		out ("\tsta\t%s\n", lbuf);
       		out ("\ttxa\n");
       		out ("\tadc\t%s+1\n", lbuf);
       		out ("\tsta\t%s+1\n", lbuf);
       		out ("\ttax\n");
		out ("\tlda\t%s\n", lbuf);
	    }
       	    break;

       	case CF_LONG:
	    if (flags & CF_CONST) {
		g_getstatic (flags, label, offs);
	     	g_inc (flags, val);
		g_putstatic (flags, label, offs);
	    } else {
		g_push (flags, 0);
		g_getstatic (flags, label, offs);
		g_add (flags, 0);
		g_putstatic (flags, label, offs);
	    }
       	    break;

       	default:
       	    typeerror (flags);
    }
}



void g_addeqlocal (unsigned flags, int offs, unsigned long val)
/* Emit += for a local variable */
{
    /* Calculate the true offset, check it, load it into Y */
    offs -= oursp;
    CheckLocalOffs (offs);

    /* Check the size and determine operation */
    switch (flags & CF_TYPE) {

       	case CF_CHAR:
       	    if (flags & CF_FORCECHAR) {
       		if (offs == 0) {
       		    out ("\tldx\t#$00\n");
       		    if (flags & CF_CONST) {
       		       	outline ("\tclc");
       	       	       	out     ("\tlda\t#$%02X\n", val & 0xFF);
       		       	outline ("\tadc\t(sp,x)");
       		       	outline ("\tsta\t(sp,x)");
       		    } else {
       		       	outline ("\tclc");
       		       	outline ("\tadc\t(sp,x)");
       		       	outline ("\tsta\t(sp,x)");
       		    }
       		} else {
       		    ldyconst (offs);
     		    out ("\tldx\t#$00\n");
     		    if (flags & CF_CONST) {
     			outline ("\tclc");
       	       	       	out     ("\tlda\t#$%02X\n", val & 0xFF);
     			outline ("\tadc\t(sp),y");
     			outline ("\tsta\t(sp),y");
     		    } else {
     		 	outline ("\tclc");
     			outline ("\tadc\t(sp),y");
     			outline ("\tsta\t(sp),y");
     		    }
     		}
     		if ((flags & CF_UNSIGNED) == 0) {
     		    out ("\tbpl\t*+3\n");
     		    out ("\tdex\n");
     		    OptAddHint ("x:!");		/* Invalidate X */
     		}
       		break;
       	    }
       	    /* FALLTHROUGH */

       	case CF_INT:
     	    if (flags & CF_CONST) {
     	     	g_getimmed (flags, val);
     	    }
     	    if (offs == 0) {
     		outline ("\tjsr\taddeq0sp");
     	    } else {
     		ldyconst (offs);
     		outline ("\tjsr\taddeqysp");
     	    }
       	    break;

       	case CF_LONG:
     	    if (flags & CF_CONST) {
	     	g_getimmed (flags, val);
	    }
	    if (offs == 0) {
		outline ("\tjsr\tladdeq0sp");
	    } else {
		ldyconst (offs);
		outline ("\tjsr\tladdeqysp");
	    }
       	    break;

       	default:
       	    typeerror (flags);
    }
}



void g_addeqind (unsigned flags, unsigned offs, unsigned long val)
/* Emit += for the location with address in ax */
{
    /* If the offset is too large for a byte register, add the high byte
     * of the offset to the primary. Beware: We need a special correction
     * if the offset in the low byte will overflow in the operation.
     */
    offs = MakeByteOffs (flags, offs);

    /* Check the size and determine operation */
    switch (flags & CF_TYPE) {

       	case CF_CHAR:
	    out ("\tsta\tptr1\n"
		 "\tstx\tptr1+1\n");
	    if (offs == 0) {
		out ("\tldx\t#$00\n"
		     "\tlda\t#$%02X\n"
		     "\tclc\n"
		     "\tadc\t(ptr1,x)\n"
		     "\tsta\t(ptr1,x)\n",
		     val & 0xFF);
	    } else {
		out ("\tldy\t#$%02X\n"
		     "\tldx\t#$00\n"
		     "\tlda\t#$%02X\n"
		     "\tclc\n"
		     "\tadc\t(ptr1),y\n"
		     "\tsta\t(ptr1),y\n",
		     offs,
		     val & 0xFF);
	    }
     	    break;

       	case CF_INT:
	    if (!FavourSize) {
		/* Lots of code, use only if size is not important */
		out ("\tsta\tptr1\n"
		     "\tstx\tptr1+1\n"
		     "\tldy\t#$%02X\n"
		     "\tlda\t#$%02X\n"
		     "\tclc\n"
       	       	     "\tadc\t(ptr1),y\n"
		     "\tsta\t(ptr1),y\n"
		     "\tpha\n"
		     "\tiny\n"
       	       	     "\tlda\t#$%02X\n"
		     "\tadc\t(ptr1),y\n"
		     "\tsta\t(ptr1),y\n"
		     "\ttax\n"
		     "\tpla\n",
		     offs,
		     val & 0xFF,
		     (val >> 8) & 0xFF);
		break;
	    }
	    /* FALL THROUGH */

       	case CF_LONG:
       	    out ("\tjsr\tpushax\n");		/* Push the address */
	    push (flags);			/* Correct the internal sp */
	    g_getind (flags, offs);		/* Fetch the value */
	    g_inc (flags, val);			/* Increment value in primary */
	    g_putind (flags, offs);		/* Store the value back */
       	    break;

       	default:
       	    typeerror (flags);
    }
}



void g_subeqstatic (unsigned flags, unsigned label, unsigned offs,
       		    unsigned long val)
/* Emit -= for a static variable */
{
    /* Create the correct label name */
    char* lbuf = GetLabelName (flags, label, offs);

    /* Check the size and determine operation */
    switch (flags & CF_TYPE) {

       	case CF_CHAR:
       	    if (flags & CF_FORCECHAR) {
       		out ("\tldx\t#$00\n");
       	  	if (flags & CF_CONST) {
       		    if (val == 1) {
       			out ("\tdec\t%s\n", lbuf);
       			out ("\tlda\t%s\n", lbuf);
       		    } else {
       		     	out ("\tsec\n");
       		     	out ("\tlda\t%s\n", lbuf);
       		     	out ("\tsbc\t#$%02X\n", val & 0xFF);
       		     	out ("\tsta\t%s\n", lbuf);
       		    }
       	  	} else {
       		    out ("\tsec\n");
       		    out ("\tsta\ttmp1\n");
       	  	    out ("\tlda\t%s\n", lbuf);
       	       	    out ("\tsbc\ttmp1\n");
       		    out ("\tsta\t%s\n", lbuf);
       	  	}
       		if ((flags & CF_UNSIGNED) == 0) {
       		    out ("\tbpl\t*+3\n");
       		    out ("\tdex\n");
       		    OptAddHint ("x:!");	       	/* Invalidate X */
       		}
       	  	break;
       	    }
       	    /* FALLTHROUGH */

       	case CF_INT:
	    outline ("\tsec");
	    if (flags & CF_CONST) {
	  	out ("\tlda\t%s\n", lbuf);
	  	out ("\tsbc\t#$%02X\n", val & 0xFF);
	  	out ("\tsta\t%s\n", lbuf);
	   	if (val < 0x100) {
	  	    label = getlabel ();
	  	    out ("\tbcs\tL%04X\n", label);
		    out ("\tdec\t%s+1\n", lbuf);
		    g_defloclabel (label);
		    out ("\tldx\t%s+1\n", lbuf);
		} else {
		    out ("\tlda\t%s+1\n", lbuf);
		    out ("\tsbc\t#$%02X\n", (val >> 8) & 0xFF);
		    out ("\tsta\t%s+1\n", lbuf);
		    out ("\ttax\n");
		    out ("\tlda\t%s\n", lbuf);
		}
	    } else {
		out ("\tsta\ttmp1\n");
		out ("\tlda\t%s\n", lbuf);
	        out ("\tsbc\ttmp1\n");
		out ("\tsta\t%s\n", lbuf);
       	       	out ("\tstx\ttmp1\n");
		out ("\tlda\t%s+1\n", lbuf);
		out ("\tsbc\ttmp1\n");
		out ("\tsta\t%s+1\n", lbuf);
		out ("\ttax\n");
		out ("\tlda\t%s\n", lbuf);
	    }
       	    break;

       	case CF_LONG:
	    if (flags & CF_CONST) {
		g_getstatic (flags, label, offs);
	     	g_dec (flags, val);
		g_putstatic (flags, label, offs);
	    } else {
		g_push (flags, 0);
		g_getstatic (flags, label, offs);
       	       	g_rsub (flags, 0);
		g_putstatic (flags, label, offs);
       	    }
       	    break;

       	default:
       	    typeerror (flags);
    }
}



void g_subeqlocal (unsigned flags, int offs, unsigned long val)
/* Emit -= for a local variable */
{
    /* Calculate the true offset, check it, load it into Y */
    offs -= oursp;
    CheckLocalOffs (offs);

    /* Check the size and determine operation */
    switch (flags & CF_TYPE) {

       	case CF_CHAR:
       	    if (flags & CF_FORCECHAR) {
    	 	ldyconst (offs);
		out ("\tldx\t#$00\n");
       	 	outline ("\tsec");
		if (flags & CF_CONST) {
		    outline ("\tlda\t(sp),y");
		    out     ("\tsbc\t#$%02X\n", val & 0xFF);
		} else {
		    outline ("\tsta\ttmp1");
		    outline ("\tlda\t(sp),y");
		    outline ("\tsbc\ttmp1");
		}
       	 	outline ("\tsta\t(sp),y");
		if ((flags & CF_UNSIGNED) == 0) {
	       	    out ("\tbpl\t*+3\n");
		    out ("\tdex\n");
		    OptAddHint ("x:!");		/* Invalidate X */
		}
       	 	break;
       	    }
       	    /* FALLTHROUGH */

       	case CF_INT:
	    if (flags & CF_CONST) {
	     	g_getimmed (flags, val);
	    }
	    if (offs == 0) {
	 	outline ("\tjsr\tsubeq0sp");
	    } else {
	 	ldyconst (offs);
	 	outline ("\tjsr\tsubeqysp");
	    }
       	    break;

       	case CF_LONG:
	    if (flags & CF_CONST) {
	     	g_getimmed (flags, val);
	    }
	    if (offs == 0) {
		outline ("\tjsr\tlsubeq0sp");
	    } else {
		ldyconst (offs);
		outline ("\tjsr\tlsubeqysp");
	    }
       	    break;

       	default:
       	    typeerror (flags);
    }
}



void g_subeqind (unsigned flags, unsigned offs, unsigned long val)
/* Emit -= for the location with address in ax */
{
    /* If the offset is too large for a byte register, add the high byte
     * of the offset to the primary. Beware: We need a special correction
     * if the offset in the low byte will overflow in the operation.
     */
    offs = MakeByteOffs (flags, offs);

    /* Check the size and determine operation */
    switch (flags & CF_TYPE) {

       	case CF_CHAR:
	    out ("\tsta\tptr1\n"
	 	 "\tstx\tptr1+1\n");
	    if (offs == 0) {
	 	out ("\tldx\t#$00\n"
	 	     "\tlda\t(ptr1,x)\n"
	 	     "\tsec\n"
	 	     "\tsbc\t#$%02X\n"
	 	     "\tsta\t(ptr1,x)\n",
	 	     val & 0xFF);
	    } else {
	 	out ("\tldy\t#$%02X\n"
	 	     "\tldx\t#$00\n"
	 	     "\tlda\t(ptr1),y\n"
	 	     "\tsec\n"
	 	     "\tsbc\t#$%02X\n"
	 	     "\tsta\t(ptr1),y\n",
	 	     offs,
	 	     val & 0xFF);
	    }
     	    break;

       	case CF_INT:
	    if (!FavourSize) {
		/* Lots of code, use only if size is not important */
		out ("\tsta\tptr1\n"
		     "\tstx\tptr1+1\n"
		     "\tldy\t#$%02X\n"
		     "\tlda\t(ptr1),y\n"
		     "\tsec\n"
		     "\tsbc\t#$%02X\n"
		     "\tsta\t(ptr1),y\n"
		     "\tpha\n"
		     "\tiny\n"
		     "\tlda\t(ptr1),y\n"
		     "\tsbc\t#$%02X\n"
		     "\tsta\t(ptr1),y\n"
		     "\ttax\n"
		     "\tpla\n",
		     offs,
		     val & 0xFF,
		     (val >> 8) & 0xFF);
		break;
	    }
	    /* FALL THROUGH */

       	case CF_LONG:
       	    out ("\tjsr\tpushax\n");		/* Push the address */
	    push (flags);  			/* Correct the internal sp */
	    g_getind (flags, offs);		/* Fetch the value */
	    g_dec (flags, val);			/* Increment value in primary */
	    g_putind (flags, offs);		/* Store the value back */
       	    break;

       	default:
       	    typeerror (flags);
    }
}



/*****************************************************************************/
/*		   Add a variable address to the value in ax		     */
/*****************************************************************************/



void g_addaddr_local (unsigned flags, int offs)
/* Add the address of a local variable to ax */
{
    /* Add the offset */
    offs -= oursp;
    if (offs != 0) {
	/* We cannot address more then 256 bytes of locals anyway */
	CheckLocalOffs (offs);
	out ("\tclc\n"
	     "\tadc\t#$%02X\n"
	     "\tbcc\t*+4\n"	/* Do also skip the CLC insn below */
	     "\tinx\n",
	     offs & 0xFF);
	/* Invalidate X */
	OptAddHint ("x:!");
    }

    /* Add the current stackpointer value */
    out ("\tclc\n"
	 "\tadc\tsp\n"
	 "\ttay\n"
	 "\ttxa\n"
	 "\tadc\tsp+1\n"
	 "\ttax\n"
	 "\ttya\n");
}



void g_addaddr_static (unsigned flags, unsigned label, unsigned offs)
/* Add the address of a static variable to ax */
{
    /* Create the correct label name */
    char* lbuf = GetLabelName (flags, label, offs);

    /* Add the address to the current ax value */
    out ("\tclc\n"
	 "\tadc\t#<(%s)\n"
	 "\ttay\n"
	 "\ttxa\n"
	 "\tadc\t#>(%s)\n"
	 "\ttax\n"
	 "\ttya\n",
	 lbuf, lbuf);
}



/*****************************************************************************/
/*				     					     */
/*****************************************************************************/



void g_save (unsigned flags)
/* Copy primary register to hold register. */
{
    /* Check the size and determine operation */
    switch (flags & CF_TYPE) {

	case CF_CHAR:
	    if (flags & CF_FORCECHAR) {
	     	outline ("\tpha");
		break;
	    }
	    /* FALLTHROUGH */

	case CF_INT:
	    outline ("\tsta\tregsave");
	    outline ("\tstx\tregsave+1");
	    break;

	case CF_LONG:
	    oljsr ("saveeax");
	    break;

	default:
	    typeerror (flags);
    }
}



void g_restore (unsigned flags)
/* Copy hold register to P. */
{
    /* Check the size and determine operation */
    switch (flags & CF_TYPE) {

	case CF_CHAR:
	    if (flags & CF_FORCECHAR) {
		outline ("\tpla");
		break;
	    }
	    /* FALLTHROUGH */

	case CF_INT:
	    outline ("\tlda\tregsave");
	    outline ("\tldx\tregsave+1");
	    break;

	case CF_LONG:
	    oljsr ("resteax");
	    break;

	default:
	    typeerror (flags);
    }
}



void g_cmp (unsigned flags, unsigned long val)
/* Immidiate compare. The primary register will not be changed, Z flag
 * will be set.
 */
{
    /* Check the size and determine operation */
    switch (flags & CF_TYPE) {

      	case CF_CHAR:
     	    if (flags & CF_FORCECHAR) {
		out ("\tcmp\t#$%02X\n", val & 0xFF);
     		break;
     	    }
     	    /* FALLTHROUGH */

     	case CF_INT:
	    out ("\tcmp\t#$%02X\n", val & 0xFF);
       	    out ("\tbne\t*+4\n");
	    out ("\tcpx\t#$%02X\n", (val >> 8) & 0xFF);
     	    break;

        case CF_LONG:
	    Internal ("g_cmp: Long compares not implemented");
	    break;

	default:
	    typeerror (flags);
    }
}



void g_immed (unsigned flags)
/* Print partial instruction to get an immediate value into the primary
 * register.
 */
{
    /* Check the size and determine operation */
    switch (flags & CF_TYPE) {

      	case CF_CHAR:
      	    if (flags & CF_FORCECHAR) {
      		outstr ("\tlda\t#");
      		break;
     	    }
     	    /* FALL THROUGH */
     	case CF_INT:
     	    outstr ("\tldax\t");
     	    break;

	case CF_LONG:
	    out ("\tjsr\tldeax\n\t.dword\t");
	    break;

	default:
	    typeerror (flags);
    }
}



static void oper (unsigned flags, unsigned long val, char** subs)
/* Encode a binary operation. subs is a pointer to four groups of three
 * strings:
 *	0-2	--> Operate on ints
 *	3-5	--> Operate on unsigneds
 *	6-8	--> Operate on longs
 *	9-11	--> Operate on unsigned longs
 *
 * The first subroutine names in each string group is used to encode an
 * operation with a zero constant, the second to encode an operation with
 * a 8 bit constant, and the third is used in all other cases.
 */
{
    unsigned offs;

    /* Determine the offset into the array */
    offs = (flags & CF_UNSIGNED)? 3 : 0;
    switch (flags & CF_TYPE) {
	case CF_CHAR:
	case CF_INT:
	    break;

	case CF_LONG:
	    offs += 6;
	    break;

	default:
	    typeerror (flags);
    }

    /* Encode the operation */
    if (flags & CF_CONST) {
	/* Constant value given */
	if (val == 0 && subs [offs+0]) {
	    /* Special case: constant with value zero */
	    oljsr (subs [offs+0]);
	} else if (val < 0x100 && subs [offs+1]) {
	    /* Special case: constant with high byte zero */
	    ldaconst (val);		/* Load low byte */
	    oljsr (subs [offs+1]);
	} else {
	    /* Others: arbitrary constant value */
	    g_getimmed (flags, val);		/* Load value */
	    oljsr (subs [offs+2]);
	}
    } else {
	/* Value not constant (is already in (e)ax) */
	oljsr (subs [offs+2]);
    }

    /* The operation will pop it's argument */
    pop (flags);
}



void g_test (unsigned flags)
/* Force a test to set cond codes right */
{
    switch (flags & CF_TYPE) {

     	case CF_CHAR:
	    if (flags & CF_FORCECHAR) {
		outline ("\ttax");
		break;
	    }
	    /* FALLTHROUGH */

     	case CF_INT:
	    outline ("\tstx\ttmp1");
	    outline ("\tora\ttmp1");
     	    break;

     	case CF_LONG:
     	    if (flags & CF_UNSIGNED) {
     		oljsr ("utsteax");
     	    } else {
     		oljsr ("tsteax");
     	    }
     	    break;

     	default:
     	    typeerror (flags);

    }
}



void g_push (unsigned flags, unsigned long val)
/* Push the primary register or a constant value onto the stack */
{
    unsigned char hi;

    if (flags & CF_CONST && (flags & CF_TYPE) != CF_LONG) {

     	/* We have a constant 8 or 16 bit value */
     	if ((flags & CF_TYPE) == CF_CHAR && (flags & CF_FORCECHAR)) {

     	    /* Handle as 8 bit value */
     	    if (FavourSize && val <= 2) {
     		out ("\tjsr\tpushc%d\n", (int) val);
     	    } else {
     		ldaconst (val);
     		oljsr ("pusha");
     	    }

     	} else {

     	    /* Handle as 16 bit value */
     	    hi = val >> 8;
     	    if (val <= 7) {
     		outstr ("\tjsr\tpush");
     		outchar (val + '0');
     		outnl ();
     	    } else if (hi == 0 || hi == 0xFF) {
     		/* Use special function */
     		ldaconst (val);
     		oljsr (hi == 0? "pusha0" : "pushaFF");
     	    } else {
     		/* Long way ... */
     		g_getimmed (flags, val);
     		oljsr ("pushax");
     	    }
     	}

    } else {

     	/* Value is not 16 bit or not constant */
     	if (flags & CF_CONST) {
     	    /* Constant 32 bit value, load into eax */
     	    g_getimmed (flags, val);
     	}

     	/* Push the primary register */
     	switch (flags & CF_TYPE) {

     	    case CF_CHAR:
     		if (flags & CF_FORCECHAR) {
     		    /* Handle as char */
     		    oljsr ("pusha");
     		    break;
     		}
     		/* FALL THROUGH */
     	    case CF_INT:
     		oljsr ("pushax");
     		break;

     	    case CF_LONG:
     		oljsr ("pusheax");
     		break;

     	    default:
     		typeerror (flags);

     	}

    }

    /* Adjust the stack offset */
    push (flags);
}



void g_swap (unsigned flags)
/* Swap the primary register and the top of the stack. flags give the type
 * of *both* values (must have same size).
 */
{
    switch (flags & CF_TYPE) {

	case CF_CHAR:
	case CF_INT:
	    oljsr ("swapstk");
	    break;

	case CF_LONG:
	    oljsr ("swapestk");
	    break;

	default:
	    typeerror (flags);

    }
}



void g_call (unsigned flags, char* lbl, unsigned argsize)
/* Call the specified subroutine name */
{
    if ((flags & CF_FIXARGC) == 0) {
	/* Pass arg count */
	ldyconst (argsize);
    }
    out ("\tjsr\t_%s\n", lbl);
    oursp += argsize;			/* callee pops args */
}



void g_callind (unsigned flags, unsigned argsize)
/* Call subroutine with address in AX */
{
    if ((flags & CF_FIXARGC) == 0) {
	/* Pass arg count */
	ldyconst (argsize);
    }
    oljsr ("callax");		/* do the call */
    oursp += argsize;		/* callee pops args */
}



void g_jump (unsigned label)
/* Jump to specified internal label number */
{
    out ("\tjmp\tL%04X\n", label);
}



void g_switch (unsigned flags)
/* Output switch statement preample */
{
    switch (flags & CF_TYPE) {

     	case CF_CHAR:
     	case CF_INT:
     	    oljsr ("switch");
     	    break;

     	case CF_LONG:
     	    oljsr ("lswitch");
     	    break;

     	default:
     	    typeerror (flags);

    }
}



void g_case (unsigned flags, unsigned label, unsigned long val)
/* Create table code for one case selector */
{
    switch (flags & CF_TYPE) {

     	case CF_CHAR:
	case CF_INT:
	    out ("\t.word\t$%04X, L%04X\n", val & 0xFFFF, label & 0xFFFF);
	    break;

	case CF_LONG:
	    out ("\t.word\t$%04X, $%04X, L%04X\n",
		 val & 0xFFFF,
		 (val >> 16) & 0xFFFF,
		 label & 0xFFFF);
	    break;

	default:
	    typeerror (flags);

    }
}



void g_truejump (unsigned flags, unsigned label)
/* Jump to label if zero flag clear */
{
    if (flags & CF_SHORT) {
	out ("\tbne\tL%04X\n", label);
    } else {
        out ("\tjne\tL%04X\n", label);
    }
}



void g_falsejump (unsigned flags, unsigned label)
/* Jump to label if zero flag set */
{
    if (flags & CF_SHORT) {
    	out ("\tbeq\tL%04X\n", label);
    } else {
       	out ("\tjeq\tL%04X\n", label);
    }
}



static void mod_internal (int k, char* verb1, char* verb2)
{
    if (k <= 8) {
	out ("\tjsr\t%ssp%c\n", verb1, k + '0');
    } else {
	CheckLocalOffs (k);
	ldyconst (k);
	out ("\tjsr\t%ssp\n", verb2);
    }
}



void g_space (int space)
/* Create or drop space on the stack */
{
    if (space < 0) {
	mod_internal (-space, "inc", "addy");
    } else if (space > 0) {
	mod_internal (space, "dec", "suby");
    }
}



void g_add (unsigned flags, unsigned long val)
/* Primary = TOS + Primary */
{
    static char* ops [12] = {
     	0,		"tosadda0",	"tosaddax",
     	0,		"tosadda0",	"tosaddax",
     	0,		0,		"tosaddeax",
     	0,		0,		"tosaddeax",
    };

    if (flags & CF_CONST) {
    	flags &= ~CF_FORCECHAR;	// Handle chars as ints
     	g_push (flags & ~CF_CONST, 0);
    }
    oper (flags, val, ops);
}



void g_sub (unsigned flags, unsigned long val)
/* Primary = TOS - Primary */
{
    static char* ops [12] = {
     	0,		"tossuba0",	"tossubax",
     	0,		"tossuba0",	"tossubax",
     	0,		0,		"tossubeax",
     	0,		0,		"tossubeax",
    };

    if (flags & CF_CONST) {
    	flags &= ~CF_FORCECHAR;	// Handle chars as ints
     	g_push (flags & ~CF_CONST, 0);
    }
    oper (flags, val, ops);
}



void g_rsub (unsigned flags, unsigned long val)
/* Primary = Primary - TOS */
{
    static char* ops [12] = {
	0,		"tosrsuba0",	"tosrsubax",
	0,		"tosrsuba0",	"tosrsubax",
	0,		0,		"tosrsubeax",
	0,		0,		"tosrsubeax",
    };
    oper (flags, val, ops);
}



void g_mul (unsigned flags, unsigned long val)
/* Primary = TOS * Primary */
{
    static char* ops [12] = {
     	0,		"tosmula0",	"tosmulax",
     	0,		"tosumula0",	"tosumulax",
     	0,		0,		"tosmuleax",
     	0,		0,		"tosumuleax",
    };

    int p2;

    /* Do strength reduction if the value is constant and a power of two */
    if (flags & CF_CONST && (p2 = powerof2 (val)) >= 0) {
     	/* Generate a shift instead */
     	g_asl (flags, p2);
	return;
    }

    /* If the right hand side is const, the lhs is not on stack but still
     * in the primary register.
     */
    if (flags & CF_CONST) {

      	switch (flags & CF_TYPE) {

      	    case CF_CHAR:
      		if (flags & CF_FORCECHAR) {
		    /* Handle some special cases */
		    switch (val) {

		     	case 3:
		     	    outline ("\tsta\ttmp1");
		     	    outline ("\tasl\ta");
		     	    outline ("\tclc");
		     	    outline ("\tadc\ttmp1");
		     	    return;

		     	case 5:
		     	    outline ("\tsta\ttmp1");
		     	    outline ("\tasl\ta");
		     	    outline ("\tasl\ta");
		     	    outline ("\tclc");
     		     	    outline ("\tadc\ttmp1");
		     	    return;

		     	case 10:
		     	    outline ("\tsta\ttmp1");
		     	    outline ("\tasl\ta");
		     	    outline ("\tasl\ta");
		     	    outline ("\tclc");
		     	    outline ("\tadc\ttmp1");
		     	    outline ("\tasl\ta");
		     	    return;
		    }
      		}
     		/* FALLTHROUGH */

	    case CF_INT:
		break;

	    case CF_LONG:
		break;

	    default:
		typeerror (flags);
	}

	/* If we go here, we didn't emit code. Push the lhs on stack and fall
	 * into the normal, non-optimized stuff.
	 */
    	flags &= ~CF_FORCECHAR;	// Handle chars as ints
     	g_push (flags & ~CF_CONST, 0);

    }

    /* Use long way over the stack */
    oper (flags, val, ops);
}



void g_div (unsigned flags, unsigned long val)
/* Primary = TOS / Primary */
{
    static char* ops [12] = {
     	0,		"tosdiva0",	"tosdivax",
     	0,		"tosudiva0",	"tosudivax",
	0,		0,		"tosdiveax",
	0,		0,		"tosudiveax",
    };

    /* Do strength reduction if the value is constant and a power of two */
    int p2;
    if ((flags & CF_CONST) && (p2 = powerof2 (val)) >= 0) {
	/* Generate a shift instead */
	g_asr (flags, p2);
    } else {
	/* Generate a division */
	if (flags & CF_CONST) {
	    /* lhs is not on stack */
    	    flags &= ~CF_FORCECHAR;	// Handle chars as ints
	    g_push (flags & ~CF_CONST, 0);
     	}
	oper (flags, val, ops);
    }
}



void g_mod (unsigned flags, unsigned long val)
/* Primary = TOS % Primary */
{
    static char* ops [12] = {
     	0,		"tosmoda0",	"tosmodax",
     	0,		"tosumoda0",	"tosumodax",
     	0,		0,		"tosmodeax",
     	0,		0,		"tosumodeax",
    };
    int p2;

    /* Check if we can do some cost reduction */
    if ((flags & CF_CONST) && (flags & CF_UNSIGNED) && val != 0xFFFFFFFF && (p2 = powerof2 (val)) >= 0) {
     	/* We can do that with an AND operation */
     	g_and (flags, val - 1);
    } else {
      	/* Do it the hard way... */
     	if (flags & CF_CONST) {
     	    /* lhs is not on stack */
    	    flags &= ~CF_FORCECHAR;	// Handle chars as ints
     	    g_push (flags & ~CF_CONST, 0);
     	}
      	oper (flags, val, ops);
    }
}



void g_or (unsigned flags, unsigned long val)
/* Primary = TOS | Primary */
{
    static char* ops [12] = {
      	0,  	     	"tosora0",	"tosorax",
      	0,  	     	"tosora0",	"tosorax",
      	0,  	     	0,		"tosoreax",
      	0,  	     	0,     		"tosoreax",
    };

    /* If the right hand side is const, the lhs is not on stack but still
     * in the primary register.
     */
    if (flags & CF_CONST) {

      	switch (flags & CF_TYPE) {

      	    case CF_CHAR:
      		if (flags & CF_FORCECHAR) {
     		    if ((val & 0xFF) != 0xFF) {
       	       	        out ("\tora\t#$%02X\n", val & 0xFF);
     		    }
      		    return;
      		}
     		/* FALLTHROUGH */

	    case CF_INT:
		if (val <= 0xFF) {
		    out ("\tora\t#$%02X\n", val & 0xFF);
		    return;
     		}
		break;

	    case CF_LONG:
		if (val <= 0xFF) {
		    out ("\tora\t#$%02X\n", val & 0xFF);
		    return;
		}
		break;

	    default:
		typeerror (flags);
	}

	/* If we go here, we didn't emit code. Push the lhs on stack and fall
	 * into the normal, non-optimized stuff.
	 */
	g_push (flags & ~CF_CONST, 0);

    }

    /* Use long way over the stack */
    oper (flags, val, ops);
}



void g_xor (unsigned flags, unsigned long val)
/* Primary = TOS ^ Primary */
{
    static char* ops [12] = {
	0,		"tosxora0",	"tosxorax",
	0,		"tosxora0",	"tosxorax",
	0,		0,	   	"tosxoreax",
	0,		0,	   	"tosxoreax",
    };


    /* If the right hand side is const, the lhs is not on stack but still
     * in the primary register.
     */
    if (flags & CF_CONST) {

      	switch (flags & CF_TYPE) {

      	    case CF_CHAR:
      		if (flags & CF_FORCECHAR) {
     		    if ((val & 0xFF) != 0) {
       	       	    	out ("\teor\t#$%02X\n", val & 0xFF);
     		    }
      		    return;
      		}
     		/* FALLTHROUGH */

	    case CF_INT:
		if (val <= 0xFF) {
		    if (val != 0) {
		     	out ("\teor\t#$%02X\n", val);
		    }
		    return;
		} else if ((val & 0xFF) == 0) {
		    out ("\tpha\n");
		    out ("\ttxa\n");
		    out	("\teor\t#$%02X\n", (val >> 8) & 0xFF);
		    out ("\ttax\n");
		    out ("\tpla");
		    return;
		}
		break;

	    case CF_LONG:
		if (val <= 0xFF) {
		    if (val != 0) {
       	       	    	out ("\teor\t#$%02X\n", val & 0xFF);
		    }
		    return;
		}
		break;

	    default:
		typeerror (flags);
	}

	/* If we go here, we didn't emit code. Push the lhs on stack and fall
	 * into the normal, non-optimized stuff.
	 */
	g_push (flags & ~CF_CONST, 0);

    }

    /* Use long way over the stack */
    oper (flags, val, ops);
}



void g_and (unsigned flags, unsigned long val)
/* Primary = TOS & Primary */
{
    static char* ops [12] = {
     	0,	     	"tosanda0",	"tosandax",
     	0,	     	"tosanda0",	"tosandax",
      	0,	     	0,		"tosandeax",
     	0,	     	0,		"tosandeax",
    };

    /* If the right hand side is const, the lhs is not on stack but still
     * in the primary register.
     */
    if (flags & CF_CONST) {

     	switch (flags & CF_TYPE) {

     	    case CF_CHAR:
     		if (flags & CF_FORCECHAR) {
     		    out ("\tand\t#$%02X\n", val & 0xFF);
     		    return;
     		}
     		/* FALLTHROUGH */
     	    case CF_INT:
		if ((val & 0xFFFF) != 0xFFFF) {
       	       	    if (val <= 0xFF) {
		    	ldxconst (0);
		    	if (val == 0) {
		    	    ldaconst (0);
		    	} else if (val != 0xFF) {
		    	    out ("\tand\t#$%02X\n", val & 0xFF);
		    	}
		    } else if ((val & 0xFF00) == 0xFF00) {
		    	out ("\tand\t#$%02X\n", val & 0xFF);
		    } else if ((val & 0x00FF) == 0x0000) {
			out ("\ttxa\n");
			out ("\tand\t#$%02X\n", (val >> 8) & 0xFF);
			out ("\ttax\n");
			ldaconst (0);
		    } else {
			out ("\ttay\n");
			out ("\ttxa\n");
			out ("\tand\t#$%02X\n", (val >> 8) & 0xFF);
			out ("\ttax\n");
			out ("\ttya\n");
			if ((val & 0x00FF) != 0x00FF) {
			    out ("\tand\t#$%02X\n", val & 0xFF);
			}
		    }
		}
		return;

	    case CF_LONG:
		if (val <= 0xFF) {
		    ldxconst (0);
		    out ("\tstx\tsreg+1\n"
		     	 "\tstx\tsreg\n");
		    if ((val & 0xFF) != 0xFF) {
		     	 out ("\tand\t#$%02X\n", val & 0xFF);
		    }
		    return;
		} else if (val == 0xFF00) {
		    ldaconst (0);
		    out ("\tsta\tsreg+1\n"
		     	 "\tsta\tsreg\n");
		    return;
		}
		break;

	    default:
		typeerror (flags);
	}

	/* If we go here, we didn't emit code. Push the lhs on stack and fall
	 * into the normal, non-optimized stuff.
	 */
	g_push (flags & ~CF_CONST, 0);

    }

    /* Use long way over the stack */
    oper (flags, val, ops);
}



void g_asr (unsigned flags, unsigned long val)
/* Primary = TOS >> Primary */
{
    static char* ops [12] = {
      	0,	     	"tosasra0",	"tosasrax",
      	0,	     	"tosshra0",	"tosshrax",
      	0,	     	0,		"tosasreax",
      	0,	     	0,		"tosshreax",
    };

    /* If the right hand side is const, the lhs is not on stack but still
     * in the primary register.
     */
    if (flags & CF_CONST) {

      	switch (flags & CF_TYPE) {

      	    case CF_CHAR:
      	    case CF_INT:
		if (val >= 1 && val <= 3) {
		    if (flags & CF_UNSIGNED) {
		       	out ("\tjsr\tshrax%ld\n", val);
		    } else {
		       	out ("\tjsr\tasrax%ld\n", val);
		    }
		    return;
		} else if (val == 8 && (flags & CF_UNSIGNED)) {
      		    out ("\ttxa\n");
      		    ldxconst (0);
		    return;
		}
		break;

	    case CF_LONG:
		if (val >= 1 && val <= 3) {
		    if (flags & CF_UNSIGNED) {
		       	out ("\tjsr\tshreax%ld\n", val);
		    } else {
		       	out ("\tjsr\tasreax%ld\n", val);
		    }
		    return;
		} else if (val == 8 && (flags & CF_UNSIGNED)) {
		    outline ("\ttxa");
		    outline ("\tldx\tsreg");
		    outline ("\tldy\tsreg+1");
		    outline ("\tsty\tsreg");
		    outline ("\tldy\t#$00");
		    outline ("\tsty\tsreg+1");
		    return;
		} else if (val == 16) {
		    outline ("\tldy\t#$00");
		    outline ("\tldx\tsreg+1");
		    if ((flags & CF_UNSIGNED) == 0) {
		        outline ("\tbpl\t*+3");
		        outline ("\tdey");
		        OptAddHint ("y:!");
		    }
     		    outline ("\tlda\tsreg");
		    outline ("\tsty\tsreg+1");
		    outline ("\tsty\tsreg");
		    return;
		}
		break;

	    default:
		typeerror (flags);
	}

	/* If we go here, we didn't emit code. Push the lhs on stack and fall
      	 * into the normal, non-optimized stuff.
	 */
	g_push (flags & ~CF_CONST, 0);

    }

    /* Use long way over the stack */
    oper (flags, val, ops);
}



void g_asl (unsigned flags, unsigned long val)
/* Primary = TOS << Primary */
{
    static char* ops [12] = {
	0,	     	"tosasla0",    	"tosaslax",
	0,	     	"tosshla0",    	"tosshlax",
	0,	     	0,     	       	"tosasleax",
	0,	     	0,     	       	"tosshleax",
    };


    /* If the right hand side is const, the lhs is not on stack but still
     * in the primary register.
     */
    if (flags & CF_CONST) {

      	switch (flags & CF_TYPE) {

      	    case CF_CHAR:
      	    case CF_INT:
		if (val >= 1 && val <= 3) {
		    if (flags & CF_UNSIGNED) {
		    	out ("\tjsr\tshlax%ld\n", val);
		    } else {
		    	out ("\tjsr\taslax%ld\n", val);
		    }
		    return;
      		} else if (val == 8) {
      		    outline ("\ttax");
      		    outline ("\tlda\t#$00");
     		    return;
     		}
     		break;

	    case CF_LONG:
		if (val >= 1 && val <= 3) {
		    if (flags & CF_UNSIGNED) {
		       	out ("\tjsr\tshleax%ld\n", val);
		    } else {
		       	out ("\tjsr\tasleax%ld\n", val);
		    }
		    return;
		} else if (val == 8) {
		    outline ("\tldy\tsreg");
		    outline ("\tsty\tsreg+1");
		    outline ("\tstx\tsreg");
		    outline ("\ttax");
		    outline ("\tlda\t#$00");
		    return;
		} else if (val == 16) {
		    outline ("\tstx\tsreg+1");
		    outline ("\tsta\tsreg");
		    outline ("\tlda\t#$00");
		    outline ("\ttax");
		    return;
		}
		break;

	    default:
		typeerror (flags);
	}

	/* If we go here, we didn't emit code. Push the lhs on stack and fall
      	 * into the normal, non-optimized stuff.
	 */
	g_push (flags & ~CF_CONST, 0);

    }

    /* Use long way over the stack */
    oper (flags, val, ops);
}



void g_neg (unsigned flags)
/* Primary = -Primary */
{
    switch (flags & CF_TYPE) {

	case CF_CHAR:
     	case CF_INT:
	    oljsr ("negax");
	    break;

	case CF_LONG:
	    oljsr ("negeax");
	    break;

	default:
	    typeerror (flags);
    }
}



void g_bneg (unsigned flags)
/* Primary = !Primary */
{
    switch (flags & CF_TYPE) {

	case CF_CHAR:
	    oljsr ("bnega");
	    break;

	case CF_INT:
	    oljsr ("bnegax");
	    break;

	case CF_LONG:
     	    oljsr ("bnegeax");
	    break;

	default:
	    typeerror (flags);
    }
}



void g_com (unsigned flags)
/* Primary = ~Primary */
{
    switch (flags & CF_TYPE) {

	case CF_CHAR:
	case CF_INT:
	    oljsr ("complax");
	    break;

	case CF_LONG:
	    oljsr ("compleax");
     	    break;

	default:
     	    typeerror (flags);
    }
}



void g_inc (unsigned flags, unsigned long val)
/* Increment the primary register by a given number */
{
    /* Don't inc by zero */
    if (val == 0) {
     	return;
    }

    /* Generate code for the supported types */
    flags &= ~CF_CONST;
    switch (flags & CF_TYPE) {

     	case CF_CHAR:
     	    if (flags & CF_FORCECHAR) {
     		out ("\tclc\n");
     		out ("\tadc\t#$%02X\n", val & 0xFF);
     		break;
     	    }
     	    /* FALLTHROUGH */

     	case CF_INT:
     	    if (FavourSize) {
     		/* Use jsr calls */
     		if (val <= 8) {
     		    out ("\tjsr\tincax%u\n", val);
     		} else if (val <= 255) {
     		    ldyconst (val);
     		    oljsr ("incaxy");
     		} else {
     		    g_add (flags | CF_CONST, val);
     		}
     	    } else {
     		/* Inline the code */
		if (val < 0x300) {
		    if ((val & 0xFF) != 0) {
			out ("\tclc\n"
			     "\tadc\t#$%02X\n"
			     "\tbcc\t*+3\n"
			     "\tinx\n",
			     (unsigned char) val);
			/* Tell the optimizer that the X register may be invalid */
			OptAddHint ("x:!");
		    }
     		    if (val >= 0x100) {
     		     	outline ("\tinx");
     		    }
     		    if (val >= 0x200) {
     			outline ("\tinx");
     		    }
     		} else {
		    outline ("\tclc");
		    if ((val & 0xFF) != 0) {
			out ("\tadc\t#$%02X\n",
			     (unsigned char) val);
			/* Tell the optimizer that the X register may be invalid */
			OptAddHint ("x:!");
		    }
     		    out ("\tpha\n"
     		     	 "\ttxa\n"
     		     	 "\tadc\t#$%02X\n"
     		     	 "\ttax\n"
     		     	 "\tpla\n",
     		     	 (unsigned char) (val >> 8));
     		}
     	    }
     	    break;

       	case CF_LONG:
     	    if (val <= 255) {
     		ldyconst (val);
     		oljsr ("inceaxy");
     	    } else {
     		g_add (flags | CF_CONST, val);
     	    }
     	    break;

     	default:
     	    typeerror (flags);

    }
}



void g_dec (unsigned flags, unsigned long val)
/* Decrement the primary register by a given number */
{
    /* Generate code for the supported types */
    flags &= ~CF_CONST;
    switch (flags & CF_TYPE) {

     	case CF_CHAR:
	    if (flags & CF_FORCECHAR) {
     		out ("\tsec\n");
		out ("\tsbc\t#$%02X\n", val & 0xFF);
		break;
     	    }
	    /* FALLTHROUGH */

     	case CF_INT:
     	    if (val <= 2) {
     		out ("\tjsr\tdecax%d\n", (int) val);
     	    } else if (val <= 255) {
		ldyconst (val);
		oljsr ("decaxy");
	    } else {
     		g_sub (flags | CF_CONST, val);
     	    }
     	    break;

     	case CF_LONG:
     	    if (val <= 255) {
     		ldyconst (val);
     		oljsr ("deceaxy");
     	    } else {
     		g_sub (flags | CF_CONST, val);
	    }
	    break;

	default:
	    typeerror (flags);

    }
}



/*
 * Following are the conditional operators. They compare the TOS against
 * the primary and put a literal 1 in the primary if the condition is
 * true, otherwise they clear the primary register
 */



void g_eq (unsigned flags, unsigned long val)
/* Test for equal */
{
    static char* ops [12] = {
     	"toseq00",	"toseqa0",	"toseqax",
     	"toseq00",	"toseqa0",	"toseqax",
     	0,		0,		"toseqeax",
     	0,		0,		"toseqeax",
    };

    /* If the right hand side is const, the lhs is not on stack but still
     * in the primary register.
     */
    if (flags & CF_CONST) {

      	switch (flags & CF_TYPE) {

      	    case CF_CHAR:
		if (flags & CF_FORCECHAR) {
		    out ("\tcmp\t#$%02X\n", val & 0xFF);
		    out ("\tjsr\tbooleq\n");
		    return;
		}
     		/* FALLTHROUGH */

      	    case CF_INT:
     		out ("\tcpx\t#$%02X\n", (val >> 8) & 0xFF);
       	       	out ("\tbne\t*+4\n");
     		out ("\tcmp\t#$%02X\n", val & 0xFF);
     		out ("\tjsr\tbooleq\n");
     		return;

     	    case CF_LONG:
     		break;

     	    default:
     		typeerror (flags);
     	}

     	/* If we go here, we didn't emit code. Push the lhs on stack and fall
      	 * into the normal, non-optimized stuff.
     	 */
     	g_push (flags & ~CF_CONST, 0);

    }

    /* Use long way over the stack */
    oper (flags, val, ops);
}



void g_ne (unsigned flags, unsigned long val)
/* Test for not equal */
{
    static char* ops [12] = {
     	"tosne00",	"tosnea0",	"tosneax",
     	"tosne00",	"tosnea0",	"tosneax",
     	0,		0,		"tosneeax",
     	0,		0,		"tosneeax",
    };


    /* If the right hand side is const, the lhs is not on stack but still
     * in the primary register.
     */
    if (flags & CF_CONST) {

      	switch (flags & CF_TYPE) {

      	    case CF_CHAR:
     		if (flags & CF_FORCECHAR) {
     		    out ("\tcmp\t#$%02X\n", val & 0xFF);
     		    out ("\tjsr\tboolne\n");
     		    return;
     		}
     		/* FALLTHROUGH */

      	    case CF_INT:
     		out ("\tcpx\t#$%02X\n", (val >> 8) & 0xFF);
     		out ("\tbne\t*+4\n");
     		out ("\tcmp\t#$%02X\n", val & 0xFF);
     		out ("\tjsr\tboolne\n");
     		return;

     	    case CF_LONG:
     		break;

     	    default:
     		typeerror (flags);
     	}

     	/* If we go here, we didn't emit code. Push the lhs on stack and fall
      	 * into the normal, non-optimized stuff.
     	 */
     	g_push (flags & ~CF_CONST, 0);

    }

    /* Use long way over the stack */
    oper (flags, val, ops);
}



void g_lt (unsigned flags, unsigned long val)
/* Test for less than */
{
    static char* ops [12] = {
     	"toslt00",	"toslta0", 	"tosltax",
     	"tosult00",	"tosulta0",	"tosultax",
     	0,		0,    	   	"toslteax",
     	0,		0,    	   	"tosulteax",
    };

    /* If the right hand side is const, the lhs is not on stack but still
     * in the primary register.
     */
    if (flags & CF_CONST) {

     	/* Give a warning in some special cases */
     	if ((flags & CF_UNSIGNED) && val == 0) {
     	    Warning (WARN_COND_NEVER_TRUE);
     	}

     	/* Look at the type */
     	switch (flags & CF_TYPE) {

     	    case CF_CHAR:
     		if (flags & CF_FORCECHAR) {
     		    out ("\tcmp\t#$%02X\n", val & 0xFF);
     		    if (flags & CF_UNSIGNED) {
     			outline ("\tjsr\tboolult");
     		    } else {
     		        outline ("\tjsr\tboollt");
     		    }
     		    return;
     		}
     		/* FALLTHROUGH */

     	    case CF_INT:
		if ((flags & CF_UNSIGNED) == 0 && val == 0) {
		    /* If we have a signed compare against zero, we only need to
		     * test the high byte.
		     */
		    out ("\ttxa\n"
			 "\tjsr\tboollt\n");
		    return;
		}
		/* Direct code only for unsigned data types */
		if (flags & CF_UNSIGNED) {
		    out ("\tcpx\t#$%02X\n"
       	       	         "\tbne\t*+4\n"
     	 	         "\tcmp\t#$%02X\n"
	 		 "\tjsr\tboolult\n",
	 		 (val >> 8) & 0xFF,
	 		 val & 0xFF);
	 	    return;
     	 	}
     	 	break;

     	    case CF_LONG:
     	 	break;

     	    default:
	 	typeerror (flags);
	}

	/* If we go here, we didn't emit code. Push the lhs on stack and fall
	 * into the normal, non-optimized stuff.
	 */
	g_push (flags & ~CF_CONST, 0);

    }

    /* Use long way over the stack */
    oper (flags, val, ops);
}



void g_le (unsigned flags, unsigned long val)
/* Test for less than or equal to */
{
    static char* ops [12] = {
	"tosle00",   	"toslea0",	"tosleax",
	"tosule00",  	"tosulea0",	"tosuleax",
	0,	     	0,    		"tosleeax",
	0,	     	0,    		"tosuleeax",
    };


    /* If the right hand side is const, the lhs is not on stack but still
     * in the primary register.
     */
    if (flags & CF_CONST) {

     	/* Look at the type */
     	switch (flags & CF_TYPE) {

	    case CF_CHAR:
		if (flags & CF_FORCECHAR) {
		    out ("\tcmp\t#$%02X\n", val & 0xFF);
		    if (flags & CF_UNSIGNED) {
		     	outline ("\tjsr\tboolule");
		    } else {
		        outline ("\tjsr\tboolle");
		    }
		    return;
		}
		/* FALLTHROUGH */

	    case CF_INT:
		if (flags & CF_UNSIGNED) {
		    out ("\tcpx\t#$%02X\n"
       	       	         "\tbne\t*+4\n"
     		         "\tcmp\t#$%02X\n"
			 "\tjsr\tboolule\n",
			 (val >> 8) & 0xFF,
			 val & 0xFF);
		    return;
		}
	  	break;

	    case CF_LONG:
		break;

	    default:
		typeerror (flags);
	}

	/* If we go here, we didn't emit code. Push the lhs on stack and fall
	 * into the normal, non-optimized stuff.
	 */
	g_push (flags & ~CF_CONST, 0);

    }

    /* Use long way over the stack */
    oper (flags, val, ops);
}



void g_gt (unsigned flags, unsigned long val)
/* Test for greater than */
{
    static char* ops [12] = {
	"tosgt00",    	"tosgta0",	"tosgtax",
	"tosugt00",   	"tosugta0",	"tosugtax",
	0,	      	0,	    	"tosgteax",
	0,	      	0, 	    	"tosugteax",
    };


    /* If the right hand side is const, the lhs is not on stack but still
     * in the primary register.
     */
    if (flags & CF_CONST) {

     	/* Look at the type */
     	switch (flags & CF_TYPE) {

	    case CF_CHAR:
		if (flags & CF_FORCECHAR) {
		    out ("\tcmp\t#$%02X\n", val & 0xFF);
		    if (flags & CF_UNSIGNED) {
		      	/* If we have a compare > 0, we will replace it by
		      	 * != 0 here, since both are identical but the latter
		      	 * is easier to optimize.
		      	 */
		      	if (val & 0xFF) {
		      	    outline ("\tjsr\tboolugt");
		      	} else {
		      	    outline ("\tjsr\tboolne");
		      	}
		    } else {
		        outline ("\tjsr\tboolgt");
		    }
		    return;
		}
		/* FALLTHROUGH */

	    case CF_INT:
		if (flags & CF_UNSIGNED) {
		    /* If we have a compare > 0, we will replace it by
		     * != 0 here, since both are identical but the latter
		     * is easier to optimize.
		     */
		    if ((val & 0xFFFF) == 0) {
			out ("\tstx\ttmp1\n"
			     "\tora\ttmp1\n"
			     "\tjsr\tboolne\n");
		    } else {
			out ("\tcpx\t#$%02X\n"
			     "\tbne\t*+4\n"
			     "\tcmp\t#$%02X\n"
			     "\tjsr\tboolugt\n",
			     (val >> 8) & 0xFF,
			     val & 0xFF);
		    }
		    return;
       	       	}
		break;

	    case CF_LONG:
		break;

	    default:
		typeerror (flags);
	}

	/* If we go here, we didn't emit code. Push the lhs on stack and fall
	 * into the normal, non-optimized stuff.
	 */
	g_push (flags & ~CF_CONST, 0);

    }

    /* Use long way over the stack */
    oper (flags, val, ops);
}



void g_ge (unsigned flags, unsigned long val)
/* Test for greater than or equal to */
{
    static char* ops [12] = {
     	"tosge00",	"tosgea0",  	"tosgeax",
     	"tosuge00",	"tosugea0",	"tosugeax",
     	0,		0,		"tosgeeax",
     	0,		0,		"tosugeeax",
    };


    /* If the right hand side is const, the lhs is not on stack but still
     * in the primary register.
     */
    if (flags & CF_CONST) {

	/* Give a warning in some special cases */
	if ((flags & CF_UNSIGNED) && val == 0) {
     	    Warning (WARN_COND_ALWAYS_TRUE);
	}

	/* Look at the type */
	switch (flags & CF_TYPE) {

	    case CF_CHAR:
		if (flags & CF_FORCECHAR) {
		    out ("\tcmp\t#$%02X\n", val & 0xFF);
		    if (flags & CF_UNSIGNED) {
			outline ("\tjsr\tbooluge");
		    } else {
		        outline ("\tjsr\tboolge");
		    }
		    return;
		}
		/* FALLTHROUGH */

	    case CF_INT:
		if (flags & CF_UNSIGNED) {
		    out ("\tcpx\t#$%02X\n"
       	       	         "\tbne\t*+4\n"
     		         "\tcmp\t#$%02X\n"
			 "\tjsr\tbooluge\n",
			 (val >> 8) & 0xFF,
			 val & 0xFF);
		    return;
		}
	  	break;

	    case CF_LONG:
		break;

	    default:
		typeerror (flags);
	}

	/* If we go here, we didn't emit code. Push the lhs on stack and fall
	 * into the normal, non-optimized stuff.
	 */
	g_push (flags & ~CF_CONST, 0);

    }

    /* Use long way over the stack */
    oper (flags, val, ops);
}



/*****************************************************************************/
/*   			   Allocating static storage			     */
/*****************************************************************************/



void g_res (unsigned n)
/* reserve static storage, n bytes */
{
    out ("\t.res\t%u\n", n);
}



void g_bytepref (void)
/* prefix for byte-sized constant */
{
    outstr ("\t.byte\t");
}



void g_wordpref (void)
/* prefix for word-sized constant */
{
    outstr ("\t.word\t");
}



void g_longpref (void)
/* Prefix for a 32 bit constant */
{
    outstr ("\t.dword\t");
}



void g_defbyte (unsigned val)
/* Define a constant byte as static data */
{
    out ("\t.byte\t$%02X\n", val & 0xFF);
}



void g_defword (unsigned val)
/* Define a constant word as static data */
{
    out ("\t.word\t$%04X\n", val & 0xFFFF);
}



void g_deflong (unsigned long val)
/* Define a constant long as static data */
{
    out ("\t.dword\t$%08lX\n", val);
}



void g_defbytes (unsigned char* bytes, unsigned count)
/* output a row of bytes as a constant */
{
    unsigned chunk;

    /* Output the stuff */
    while (count) {

	/* How many go into this line? */
	if ((chunk = count) > 16) {
	    chunk = 16;
	}
	count -= chunk;

	/* Output one line */
	g_bytepref ();
	do {
	    out ("$%02X", *bytes++ & 0xFF);
	    if (--chunk) {
		outchar (',');
	    }
	} while (chunk);
	outchar ('\n');

        /* keep output buf from overflowing */
        flushout ();
    }
}



void g_zerobytes (unsigned n)
/* Output n bytes of data initialized with zero */
{
    out ("\t.res\t%u\n", n);
}



/*****************************************************************************/
/*			    Inlined known functions			     */
/*****************************************************************************/



void g_strlen (unsigned flags, unsigned val, unsigned offs)
/* Inline the strlen() function */
{
    /* We need a label in both cases */
    unsigned label = getlabel ();

    /* Two different encodings */
    if (flags & CF_CONST) {

	/* The address of the string is constant. Create the correct label name */
    	char* lbuf = GetLabelName (flags, val, offs);

	/* Generate the strlen code */
	out ("\tldx\t#$00\n");
	out ("\tldy\t#$FF\n");
	g_defloclabel (label);
	out ("\tiny\n");
	out ("\tlda\t%s,y\n", lbuf);
	out ("\tbne\tL%04X\n", label);
	out ("\ttya\n");

    } else {

       	/* Address not constant but in primary */
	if (FavourSize) {
	    /* This is too much code, so call strlen instead of inlining */
    	    out ("\tjsr\t_strlen\n");
	} else {
	    /* Inline the function */
	    out ("\tsta\tptr1\n");
	    out ("\tstx\tptr1+1\n");
	    out ("\tldx\t#$00\n");
	    out ("\tldy\t#$FF\n");
	    g_defloclabel (label);
	    out ("\tiny\n");
	    out ("\tlda\t(ptr1),y\n");
	    out ("\tbne\tL%04X\n", label);
	    out ("\ttya\n");
     	}
    }
}



